# Copyright (C) 2023  Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
"""Exceptions for the fastkml package."""


class FastKMLError(Exception):
    """Base class for all fastkml exceptions."""


class KMLParseError(FastKMLError):
    """Raised when there is an error parsing KML."""


class KMLWriteError(FastKMLError):
    """Raised when there is an error writing KML."""


class KMLSchemaError(FastKMLError):
    """Raised when there is an error with the KML Schema."""


# geometry and kml_coordinates are mutually exclusive
class GeometryError(FastKMLError):
    """Raised when there is an error with the geometry."""



# Copyright (C) 2023  Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
"""Container classes for KML elements."""
import logging
import urllib.parse as urlparse
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Union

from fastkml import atom
from fastkml import gx
from fastkml.data import ExtendedData
from fastkml.data import Schema
from fastkml.features import Placemark
from fastkml.features import Snippet
from fastkml.features import _Feature
from fastkml.geometry import LinearRing
from fastkml.geometry import LineString
from fastkml.geometry import MultiGeometry
from fastkml.geometry import Point
from fastkml.geometry import Polygon
from fastkml.helpers import xml_subelement_list
from fastkml.helpers import xml_subelement_list_kwarg
from fastkml.registry import RegistryItem
from fastkml.registry import registry
from fastkml.styles import Style
from fastkml.styles import StyleMap
from fastkml.styles import StyleUrl
from fastkml.times import TimeSpan
from fastkml.times import TimeStamp
from fastkml.utils import find_all
from fastkml.views import Camera
from fastkml.views import LookAt
from fastkml.views import Region

logger = logging.getLogger(__name__)

KmlGeometry = Union[
    Point,
    LineString,
    LinearRing,
    Polygon,
    MultiGeometry,
    gx.MultiTrack,
    gx.Track,
]


class _Container(_Feature):
    """
    A Container element that holds one or more Features.

    Supports the creation of nested hierarchies.
    subclasses are:
    Document,
    Folder.
    """

    features: List[_Feature]

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        name: Optional[str] = None,
        visibility: Optional[bool] = None,
        isopen: Optional[bool] = None,
        atom_link: Optional[atom.Link] = None,
        atom_author: Optional[atom.Author] = None,
        address: Optional[str] = None,
        phone_number: Optional[str] = None,
        snippet: Optional[Snippet] = None,
        description: Optional[str] = None,
        view: Optional[Union[Camera, LookAt]] = None,
        times: Optional[Union[TimeSpan, TimeStamp]] = None,
        style_url: Optional[StyleUrl] = None,
        styles: Optional[Iterable[Union[Style, StyleMap]]] = None,
        region: Optional[Region] = None,
        extended_data: Optional[ExtendedData] = None,
        # Container specific
        features: Optional[Iterable[_Feature]] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            id=id,
            target_id=target_id,
            name=name,
            visibility=visibility,
            isopen=isopen,
            atom_link=atom_link,
            atom_author=atom_author,
            address=address,
            phone_number=phone_number,
            snippet=snippet,
            description=description,
            view=view,
            times=times,
            style_url=style_url,
            styles=styles,
            region=region,
            extended_data=extended_data,
            **kwargs,
        )
        self.features = list(features) if features else []

    def __repr__(self) -> str:
        """Create a string (c)representation for _Container."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"target_id={self.target_id!r}, "
            f"name={self.name!r}, "
            f"visibility={self.visibility!r}, "
            f"isopen={self.isopen!r}, "
            f"atom_link={self.atom_link!r}, "
            f"atom_author={self.atom_author!r}, "
            f"address={self.address!r}, "
            f"phone_number={self.phone_number!r}, "
            f"snippet={self.snippet!r}, "
            f"description={self.description!r}, "
            f"view={self.view!r}, "
            f"times={self.times!r}, "
            f"style_url={self.style_url!r}, "
            f"styles={self.styles!r}, "
            f"region={self.region!r}, "
            f"extended_data={self.extended_data!r}, "
            f"features={self.features!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    def append(self, kmlobj: _Feature) -> None:
        """Append a feature."""
        if kmlobj is self:
            msg = "Cannot append self"
            raise ValueError(msg)
        assert self.features is not None  # noqa: S101
        self.features.append(kmlobj)


class Folder(_Container):
    """
    A Folder is used to arrange other Features hierarchically.

    It may contain Folders, Placemarks, NetworkLinks, or Overlays.
    """


class Document(_Container):
    """
    A Document is a container for features and styles.

    This element is required if your KML file uses shared styles or schemata for typed
    extended data.
    """

    schemata: List[Schema]

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        name: Optional[str] = None,
        visibility: Optional[bool] = None,
        isopen: Optional[bool] = None,
        atom_link: Optional[atom.Link] = None,
        atom_author: Optional[atom.Author] = None,
        address: Optional[str] = None,
        phone_number: Optional[str] = None,
        snippet: Optional[Snippet] = None,
        description: Optional[str] = None,
        view: Optional[Union[Camera, LookAt]] = None,
        times: Optional[Union[TimeSpan, TimeStamp]] = None,
        style_url: Optional[StyleUrl] = None,
        styles: Optional[Iterable[Union[Style, StyleMap]]] = None,
        region: Optional[Region] = None,
        extended_data: Optional[ExtendedData] = None,
        features: Optional[List[_Feature]] = None,
        schemata: Optional[Iterable[Schema]] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a new instance of the class.

        Args:
        ----
            ns (Optional[str]): The namespace.
            name_spaces (Optional[Dict[str, str]]):
                The dictionary of namespace prefixes and URIs.
            id (Optional[str]): The ID of the container.
            target_id (Optional[str]): The target ID.
            name (Optional[str]): The name of the container.
            visibility (Optional[bool]): The visibility flag.
            isopen (Optional[bool]): The isopen flag.
            atom_link (Optional[atom.Link]): The Atom link.
            atom_author (Optional[atom.Author]): The Atom author.
            address (Optional[str]): The address.
            phone_number (Optional[str]): The phone number.
            snippet (Optional[Snippet]): The snippet.
            description (Optional[str]): The description.
            view (Optional[Union[Camera, LookAt]]): The view.
            times (Optional[Union[TimeSpan, TimeStamp]]): The times.
            style_url (Optional[StyleUrl]): The style URL.
            styles (Optional[Iterable[Union[Style, StyleMap]]]): The styles.
            region (Optional[Region]): The region.
            extended_data (Optional[ExtendedData]): The extended data.
            features (Optional[List[_Feature]]): The list of features.
            schemata (Optional[Iterable[Schema]]): The schemata.
            **kwargs (Any): Additional keyword arguments.

        Returns:
        -------
            None

        """
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            id=id,
            target_id=target_id,
            name=name,
            visibility=visibility,
            isopen=isopen,
            atom_link=atom_link,
            atom_author=atom_author,
            address=address,
            phone_number=phone_number,
            snippet=snippet,
            description=description,
            view=view,
            times=times,
            style_url=style_url,
            styles=styles,
            region=region,
            extended_data=extended_data,
            features=features,
            **kwargs,
        )
        self.schemata = list(schemata) if schemata else []

    def __repr__(self) -> str:
        """Create a string (c)representation for Document."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"target_id={self.target_id!r}, "
            f"name={self.name!r}, "
            f"visibility={self.visibility!r}, "
            f"isopen={self.isopen!r}, "
            f"atom_link={self.atom_link!r}, "
            f"atom_author={self.atom_author!r}, "
            f"address={self.address!r}, "
            f"phone_number={self.phone_number!r}, "
            f"snippet={self.snippet!r}, "
            f"description={self.description!r}, "
            f"view={self.view!r}, "
            f"times={self.times!r}, "
            f"style_url={self.style_url!r}, "
            f"styles={self.styles!r}, "
            f"region={self.region!r}, "
            f"extended_data={self.extended_data!r}, "
            f"features={self.features!r}, "
            f"schemata={self.schemata!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    def get_style_by_url(self, style_url: str) -> Optional[Union[Style, StyleMap]]:
        """
        Get a style by URL.

        Parameters
        ----------
        style_url : str
            The URL of the style.

        Returns
        -------
        Optional[Union[Style, StyleMap]]
            The style object if found, otherwise None.

        """
        id_ = urlparse.urlparse(style_url).fragment
        return next(
            find_all(  # type: ignore[arg-type]
                self,
                of_type=(Style, StyleMap),
                id=id_,
            ),
            None,
        )


registry.register(
    _Container,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="features",
        node_name="Folder,Placemark,Document",
        classes=(Folder, Placemark, Document),
        get_kwarg=xml_subelement_list_kwarg,
        set_element=xml_subelement_list,
    ),
)
registry.register(
    Document,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="schemata",
        node_name="Schema",
        classes=(Schema,),
        get_kwarg=xml_subelement_list_kwarg,
        set_element=xml_subelement_list,
    ),
)



# Copyright (C) 2023  Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
"""KML Views."""
import logging
from typing import Any
from typing import Dict
from typing import Optional

from fastkml import config
from fastkml.base import _XMLObject
from fastkml.enums import AltitudeMode
from fastkml.helpers import enum_subelement
from fastkml.helpers import float_subelement
from fastkml.helpers import int_subelement
from fastkml.helpers import subelement_enum_kwarg
from fastkml.helpers import subelement_float_kwarg
from fastkml.helpers import subelement_int_kwarg
from fastkml.helpers import xml_subelement
from fastkml.helpers import xml_subelement_kwarg
from fastkml.kml_base import _BaseObject
from fastkml.mixins import TimeMixin
from fastkml.registry import RegistryItem
from fastkml.registry import registry

__all__ = ["Camera", "LatLonAltBox", "Lod", "LookAt", "Region"]

logger = logging.getLogger(__name__)


class _AbstractView(TimeMixin, _BaseObject):
    """
    Abstract element that cannot be used directly in a KML file.

    This element is extended by the <Camera> and <LookAt> elements.
    """

    longitude: Optional[float]
    # Longitude of the virtual camera (eye point). Angular distance in degrees,
    # relative to the Prime Meridian. Values west of the Meridian range from
    # -180 to 0 degrees. Values east of the Meridian range from 0 to 180 degrees.

    latitude: Optional[float]
    # Latitude of the virtual camera. Degrees north or south of the Equator
    # (0 degrees). Values range from -90 degrees to 90 degrees.

    altitude: Optional[float]
    # Distance of the camera from the earth's surface, in meters. Interpreted
    # according to the Camera's <altitudeMode> or <gx:altitudeMode>.

    heading: Optional[float]
    # Direction (azimuth) of the camera, in degrees. Default=0 (true North).
    # (See diagram.) Values range from 0 to 360 degrees.

    tilt: Optional[float]
    # Rotation, in degrees, of the camera around the X axis. A value of 0
    # indicates that the view is aimed straight down toward the earth (the
    # most common case). A value for 90 for <tilt> indicates that the view
    # is aimed toward the horizon. Values greater than 90 indicate that the
    # view is pointed up into the sky. Values for <tilt> are clamped at +180
    # degrees.

    altitude_mode: Optional[AltitudeMode]
    # Specifies how the <altitude> specified for the Camera is interpreted.
    # Possible values are as follows:
    #   relativeToGround -
    #       (default) Interprets the <altitude> as a value in meters above the
    #       ground. If the point is over water, the <altitude> will be
    #       interpreted as a value in meters above sea level. See
    #       <gx:altitudeMode> below to specify points relative to the sea floor.
    #   clampToGround -
    #       For a camera, this setting also places the camera relativeToGround,
    #       since putting the camera exactly at terrain height would mean that
    #       the eye would intersect the terrain (and the view would be blocked).
    #   absolute -
    #       Interprets the <altitude> as a value in meters above sea level.

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        longitude: Optional[float] = None,
        latitude: Optional[float] = None,
        altitude: Optional[float] = None,
        heading: Optional[float] = None,
        tilt: Optional[float] = None,
        altitude_mode: Optional[AltitudeMode] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a new View object.

        Parameters
        ----------
        ns : Optional[str]
            The namespace for the view.
        name_spaces : Optional[Dict[str, str]]
            The dictionary of namespace prefixes and URIs.
        id : Optional[str]
            The ID of the view.
        target_id : Optional[str]
            The ID of the target view.
        longitude : Optional[float]
            The longitude coordinate of the view.
        latitude : Optional[float]
            The latitude coordinate of the view.
        altitude : Optional[float]
            The altitude coordinate of the view.
        heading : Optional[float]
            The heading angle of the view.
        tilt : Optional[float]
            The tilt angle of the view.
        altitude_mode : Optional[AltitudeMode]
            The altitude mode of the view.
        kwargs : Any
            Additional keyword arguments.

        Returns
        -------
            None

        """
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            id=id,
            target_id=target_id,
            **kwargs,
        )
        self.longitude = longitude
        self.latitude = latitude
        self.altitude = altitude
        self.heading = heading
        self.tilt = tilt
        self.altitude_mode = altitude_mode

    def __repr__(self) -> str:
        """Create a string (c)representation for _AbstractView."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"target_id={self.target_id!r}, "
            f"longitude={self.longitude!r}, "
            f"latitude={self.latitude!r}, "
            f"altitude={self.altitude!r}, "
            f"heading={self.heading!r}, "
            f"tilt={self.tilt!r}, "
            f"altitude_mode={self.altitude_mode}, "
            f"**{self._get_splat()!r},"
            ")"
        )


registry.register(
    _AbstractView,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="longitude",
        node_name="longitude",
        classes=(float,),
        get_kwarg=subelement_float_kwarg,
        set_element=float_subelement,
        default=0.0,
    ),
)
registry.register(
    _AbstractView,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="latitude",
        node_name="latitude",
        classes=(float,),
        get_kwarg=subelement_float_kwarg,
        set_element=float_subelement,
        default=0.0,
    ),
)
registry.register(
    _AbstractView,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="altitude",
        node_name="altitude",
        classes=(float,),
        get_kwarg=subelement_float_kwarg,
        set_element=float_subelement,
        default=0.0,
    ),
)
registry.register(
    _AbstractView,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="heading",
        node_name="heading",
        classes=(float,),
        get_kwarg=subelement_float_kwarg,
        set_element=float_subelement,
        default=0.0,
    ),
)
registry.register(
    _AbstractView,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="tilt",
        node_name="tilt",
        classes=(float,),
        get_kwarg=subelement_float_kwarg,
        set_element=float_subelement,
        default=0.0,
    ),
)


class Camera(_AbstractView):
    """
    Defines the virtual camera that views the scene.

    This element defines the position of the camera relative to the Earth's surface
    as well as the viewing direction of the camera. The camera position is defined
    by <longitude>, <latitude>, <altitude>, and either <altitudeMode> or
    <gx:altitudeMode>. The viewing direction of the camera is defined by
    <heading>, <tilt>, and <roll>. <Camera> can be a child element of any
    Feature or of <NetworkLinkControl>. A parent element cannot contain both a
    <Camera> and a <LookAt> at the same time.

    <Camera> provides full six-degrees-of-freedom control over the view,
    so you can position the Camera in space and then rotate it around the
    X, Y, and Z axes. Most importantly, you can tilt the camera view so that
    you're looking above the horizon into the sky.

    <Camera> can also contain a TimePrimitive (<gx:TimeSpan> or <gx:TimeStamp>).
    Time values in Camera affect historical imagery, sunlight, and the display of
    time-stamped features. For more information, read Time with AbstractViews in
    the Time and Animation chapter of the Developer's Guide.

    https://developers.google.com/kml/documentation/kmlreference#camera
    """

    roll: Optional[float]
    # Rotation, in degrees, of the camera around the Z axis. Values range from
    # -180 to +180 degrees.

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        longitude: Optional[float] = None,
        latitude: Optional[float] = None,
        altitude: Optional[float] = None,
        heading: Optional[float] = None,
        tilt: Optional[float] = None,
        roll: Optional[float] = None,
        altitude_mode: Optional[AltitudeMode] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a new View object.

        Args:
        ----
            ns (Optional[str]): The namespace for the view.
            name_spaces (Optional[Dict[str, str]]): The dictionary of namespaces.
            id (Optional[str]): The ID of the view.
            target_id (Optional[str]): The target ID of the view.
            longitude (Optional[float]): The longitude of the view.
            latitude (Optional[float]): The latitude of the view.
            altitude (Optional[float]): The altitude of the view.
            heading (Optional[float]): The heading of the view.
            tilt (Optional[float]): The tilt of the view.
            roll (Optional[float]): The roll of the view.
            altitude_mode (AltitudeMode): The altitude mode of the view.
            **kwargs (Any): Additional keyword arguments.

        Returns:
        -------
            None

        """
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            id=id,
            target_id=target_id,
            longitude=longitude,
            latitude=latitude,
            altitude=altitude,
            heading=heading,
            tilt=tilt,
            altitude_mode=altitude_mode,
            **kwargs,
        )
        self.roll = roll

    def __repr__(self) -> str:
        """Create a string (c)representation for Camera."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"target_id={self.target_id!r}, "
            f"longitude={self.longitude!r}, "
            f"latitude={self.latitude!r}, "
            f"altitude={self.altitude!r}, "
            f"heading={self.heading!r}, "
            f"tilt={self.tilt!r}, "
            f"roll={self.roll!r}, "
            f"altitude_mode={self.altitude_mode}, "
            f"**{self._get_splat()!r},"
            ")"
        )


registry.register(
    Camera,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="roll",
        node_name="roll",
        classes=(float,),
        get_kwarg=subelement_float_kwarg,
        set_element=float_subelement,
        default=0.0,
    ),
)
registry.register(
    Camera,
    RegistryItem(
        ns_ids=("kml", "gx", ""),
        attr_name="altitude_mode",
        node_name="altitudeMode",
        classes=(AltitudeMode,),
        get_kwarg=subelement_enum_kwarg,
        set_element=enum_subelement,
        default=AltitudeMode.clamp_to_ground,
    ),
)


class LookAt(_AbstractView):
    """
    Defines a virtual camera that is associated with any element derived from Feature.

    The LookAt element positions the "camera" in relation to the object that is being
    viewed. In Google Earth, the view "flies to" this LookAt viewpoint when the user
    double-clicks an item in the Places panel or double-clicks an icon in the 3D viewer.

    https://developers.google.com/kml/documentation/kmlreference#lookat
    """

    range: Optional[float]
    # Distance in meters from the point specified by <longitude>, <latitude>,
    # and <altitude> to the LookAt position.

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        longitude: Optional[float] = None,
        latitude: Optional[float] = None,
        altitude: Optional[float] = None,
        heading: Optional[float] = None,
        tilt: Optional[float] = None,
        range: Optional[float] = None,
        altitude_mode: Optional[AltitudeMode] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a new instance of the class.

        Args:
        ----
            ns (Optional[str]): The namespace for the element.
            name_spaces (Optional[Dict[str, str]]): The dictionary of namespaces.
            id (Optional[str]): The ID of the element.
            target_id (Optional[str]): The target ID of the element.
            longitude (Optional[float]): The longitude value.
            latitude (Optional[float]): The latitude value.
            altitude (Optional[float]): The altitude value.
            heading (Optional[float]): The heading value.
            tilt (Optional[float]): The tilt value.
            range (Optional[float]): The range value.
            altitude_mode (AltitudeMode): The altitude mode. Defaults to
                AltitudeMode.relative_to_ground.
            **kwargs (Any): Additional keyword arguments.

        Returns:
        -------
            None

        """
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            id=id,
            target_id=target_id,
            longitude=longitude,
            latitude=latitude,
            altitude=altitude,
            heading=heading,
            tilt=tilt,
            altitude_mode=altitude_mode,
            **kwargs,
        )
        self.range = range

    def __repr__(self) -> str:
        """Create a string (c)representation for LookAt."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"target_id={self.target_id!r}, "
            f"longitude={self.longitude!r}, "
            f"latitude={self.latitude!r}, "
            f"altitude={self.altitude!r}, "
            f"heading={self.heading!r}, "
            f"tilt={self.tilt!r}, "
            f"range={self.range!r}, "
            f"altitude_mode={self.altitude_mode}, "
            f"**{self._get_splat()!r},"
            ")"
        )


registry.register(
    LookAt,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="range",
        node_name="range",
        classes=(float,),
        get_kwarg=subelement_float_kwarg,
        set_element=float_subelement,
    ),
)
registry.register(
    LookAt,
    RegistryItem(
        ns_ids=("kml", "gx", ""),
        attr_name="altitude_mode",
        node_name="altitudeMode",
        classes=(AltitudeMode,),
        get_kwarg=subelement_enum_kwarg,
        set_element=enum_subelement,
        default=AltitudeMode.clamp_to_ground,
    ),
)


class LatLonAltBox(_XMLObject):
    """
    A bounding box defined by geographic coordinates and altitudes.

    https://developers.google.com/kml/documentation/kmlreference#latlonaltbox
    """

    _default_nsid = config.KML

    north: Optional[float]
    south: Optional[float]
    east: Optional[float]
    west: Optional[float]
    min_altitude: Optional[float]
    max_altitude: Optional[float]
    altitude_mode: Optional[AltitudeMode]

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        north: Optional[float] = None,
        south: Optional[float] = None,
        east: Optional[float] = None,
        west: Optional[float] = None,
        min_altitude: Optional[float] = None,
        max_altitude: Optional[float] = None,
        altitude_mode: Optional[AltitudeMode] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a new View object.

        Args:
        ----
            ns (Optional[str]): The namespace for the view.
            name_spaces (Optional[Dict[str, str]]): A dictionary of namespace prefixes
                and URIs.
            north (Optional[float]): The northern latitude of the view.
            south (Optional[float]): The southern latitude of the view.
            east (Optional[float]): The eastern longitude of the view.
            west (Optional[float]): The western longitude of the view.
            min_altitude (Optional[float]): The minimum altitude of the view.
            max_altitude (Optional[float]): The maximum altitude of the view.
            altitude_mode (Optional[AltitudeMode]): The altitude mode of the view.
            **kwargs (Any): Additional keyword arguments.

        Returns:
        -------
            None

        """
        super().__init__(ns=ns, name_spaces=name_spaces, **kwargs)
        self.north = north
        self.south = south
        self.east = east
        self.west = west
        self.min_altitude = min_altitude
        self.max_altitude = max_altitude
        self.altitude_mode = altitude_mode

    def __repr__(self) -> str:
        """Create a string (c)representation for LatLonAltBox."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"north={self.north!r}, "
            f"south={self.south!r}, "
            f"east={self.east!r}, "
            f"west={self.west!r}, "
            f"min_altitude={self.min_altitude!r}, "
            f"max_altitude={self.max_altitude!r}, "
            f"altitude_mode={self.altitude_mode}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    def __bool__(self) -> bool:
        """
        Check if all the attributes (north, south, east, west) are not None.

        Returns
        -------
            bool: True if all attributes are not None, False otherwise.

        """
        return all(
            (
                self.north is not None,
                self.south is not None,
                self.east is not None,
                self.west is not None,
            ),
        )


registry.register(
    LatLonAltBox,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="north",
        node_name="north",
        classes=(float,),
        get_kwarg=subelement_float_kwarg,
        set_element=float_subelement,
    ),
)
registry.register(
    LatLonAltBox,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="south",
        node_name="south",
        classes=(float,),
        get_kwarg=subelement_float_kwarg,
        set_element=float_subelement,
    ),
)
registry.register(
    LatLonAltBox,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="east",
        node_name="east",
        classes=(float,),
        get_kwarg=subelement_float_kwarg,
        set_element=float_subelement,
    ),
)
registry.register(
    LatLonAltBox,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="west",
        node_name="west",
        classes=(float,),
        get_kwarg=subelement_float_kwarg,
        set_element=float_subelement,
    ),
)
registry.register(
    LatLonAltBox,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="min_altitude",
        node_name="minAltitude",
        classes=(float,),
        get_kwarg=subelement_float_kwarg,
        set_element=float_subelement,
        default=0.0,
    ),
)
registry.register(
    LatLonAltBox,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="max_altitude",
        node_name="maxAltitude",
        classes=(float,),
        get_kwarg=subelement_float_kwarg,
        set_element=float_subelement,
        default=0.0,
    ),
)
registry.register(
    LatLonAltBox,
    RegistryItem(
        ns_ids=("kml", "gx", ""),
        attr_name="altitude_mode",
        node_name="altitudeMode",
        classes=(AltitudeMode,),
        get_kwarg=subelement_enum_kwarg,
        set_element=enum_subelement,
        default=AltitudeMode.clamp_to_ground,
    ),
)


class Lod(_XMLObject):
    """
    Lod is an abbreviation for Level of Detail.

    <Lod> describes the size of the projected region on the screen that is required in
    order for the region to be considered "active."
    Also specifies the size of the pixel ramp used for fading in
    (from transparent to opaque) and fading out (from opaque to transparent).

    https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-region
    """

    _default_nsid = config.KML

    min_lod_pixels: Optional[int]
    max_lod_pixels: Optional[int]
    min_fade_extent: Optional[int]
    max_fade_extent: Optional[int]

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        min_lod_pixels: Optional[int] = None,
        max_lod_pixels: Optional[int] = None,
        min_fade_extent: Optional[int] = None,
        max_fade_extent: Optional[int] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a new instance of the View class.

        Args:
        ----
            ns (Optional[str]): The namespace for the view.
            name_spaces (Optional[Dict[str, str]]): The dictionary of namespace prefixes
                and URIs.
            min_lod_pixels (Optional[int]): The minimum level of detail in pixels.
            max_lod_pixels (Optional[int]): The maximum level of detail in pixels.
            min_fade_extent (Optional[int]): The minimum fade extent in pixels.
            max_fade_extent (Optional[int]): The maximum fade extent in pixels.
            **kwargs (Any): Additional keyword arguments.

        Returns:
        -------
            None

        """
        super().__init__(ns=ns, name_spaces=name_spaces, **kwargs)
        self.min_lod_pixels = min_lod_pixels
        self.max_lod_pixels = max_lod_pixels
        self.min_fade_extent = min_fade_extent
        self.max_fade_extent = max_fade_extent

    def __repr__(self) -> str:
        """Create a string (c)representation for Lod."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"min_lod_pixels={self.min_lod_pixels!r}, "
            f"max_lod_pixels={self.max_lod_pixels!r}, "
            f"min_fade_extent={self.min_fade_extent!r}, "
            f"max_fade_extent={self.max_fade_extent!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    def __bool__(self) -> bool:
        """
        Check if the object is considered True or False.

        Returns
        -------
            bool: True if the `min_lod_pixels` attribute is not None, False otherwise.

        """
        return self.min_lod_pixels is not None


registry.register(
    Lod,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="min_lod_pixels",
        node_name="minLodPixels",
        classes=(float,),
        get_kwarg=subelement_int_kwarg,
        set_element=int_subelement,
        default=256,
    ),
)
registry.register(
    Lod,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="max_lod_pixels",
        node_name="maxLodPixels",
        classes=(float,),
        get_kwarg=subelement_int_kwarg,
        set_element=int_subelement,
        default=-1,
    ),
)
registry.register(
    Lod,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="min_fade_extent",
        node_name="minFadeExtent",
        classes=(float,),
        get_kwarg=subelement_int_kwarg,
        set_element=int_subelement,
        default=0,
    ),
)
registry.register(
    Lod,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="max_fade_extent",
        node_name="maxFadeExtent",
        classes=(float,),
        get_kwarg=subelement_int_kwarg,
        set_element=int_subelement,
        default=0,
    ),
)


class Region(_BaseObject):
    """
    A region contains a bounding box that describes an area of interest.

    In addition, a Region contains an LOD (level of detail) extent (<Lod>),
    which is a pair of projected coordinate bounding boxes that describe
    the area that should be loaded in the viewport corresponding to a given
    level of detail.

    https://developers.google.com/kml/documentation/kmlreference#region
    """

    lat_lon_alt_box: Optional[LatLonAltBox]
    lod: Optional[Lod]

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        lat_lon_alt_box: Optional[LatLonAltBox] = None,
        lod: Optional[Lod] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a new instance of the View class.

        Args:
        ----
        ns : Optional[str]
            The namespace of the view.
        name_spaces : Optional[Dict[str, str]]
            The dictionary of namespace prefixes and URIs.
        id : Optional[str]
            The ID of the view.
        target_id : Optional[str]
            The target ID of the view.
        lat_lon_alt_box : Optional[LatLonAltBox]
            The latitude, longitude, and altitude box of the view.
        lod : Optional[Lod]
            The level of detail of the view.
        **kwargs : Any
            Additional keyword arguments.

        Returns:
        -------
        None

        """
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            id=id,
            target_id=target_id,
            **kwargs,
        )
        self.lat_lon_alt_box = lat_lon_alt_box
        self.lod = lod

    def __repr__(self) -> str:
        """Create a string (c)representation for Region."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"target_id={self.target_id!r}, "
            f"lat_lon_alt_box={self.lat_lon_alt_box!r}, "
            f"lod={self.lod!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    def __bool__(self) -> bool:
        """
        Check if the object is considered True or False.

        Returns
        -------
            bool: True if the `lat_lon_alt_box` attribute is not empty, False otherwise.

        """
        return bool(self.lat_lon_alt_box)


registry.register(
    Region,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="lat_lon_alt_box",
        node_name="LatLonAltBox",
        classes=(LatLonAltBox,),
        get_kwarg=xml_subelement_kwarg,
        set_element=xml_subelement,
    ),
)
registry.register(
    Region,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="lod",
        node_name="Lod",
        classes=(Lod,),
        get_kwarg=xml_subelement_kwarg,
        set_element=xml_subelement,
    ),
)



# Copyright (C) 2024 Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
"""Validate KML files against the XML schema."""

import logging
import pathlib
from functools import lru_cache
from typing import Final
from typing import Optional

from fastkml import config
from fastkml.types import Element

logger = logging.getLogger(__name__)

MUTUAL_EXCLUSIVE: Final = "Only one of element and file_to_validate can be provided."
REQUIRE_ONE_OF: Final = "Either element or file_to_validate must be provided."


@lru_cache(maxsize=16)
def get_schema_parser(
    schema: Optional[pathlib.Path] = None,
) -> "config.etree.XMLSchema":
    """
    Parse the XML schema.

    Args:
    ----
        schema: The path to the XML schema file.

    Returns:
    -------
        The parsed XML schema.

    To clear the cache call get_schema_parser.cache_clear().

    """
    if schema is None:
        schema = pathlib.Path(__file__).parent / "schema" / "ogckml22.xsd"
    return config.etree.XMLSchema(config.etree.parse(schema))


def validate(
    *,
    schema: Optional[pathlib.Path] = None,
    element: Optional[Element] = None,
    file_to_validate: Optional[pathlib.Path] = None,
) -> Optional[bool]:
    """
    Validate a KML file against the XML schema.

    Args:
    ----
        schema: The path to the XML schema file.
        element: The element to validate.
        file_to_validate: The file to validate.

    Returns:
    -------
            True if the file or element is valid.
            Raises an AssertionError if validation fails.
            Returns None if the schema parser is unavailable.

    """
    if element is None and file_to_validate is None:
        raise ValueError(REQUIRE_ONE_OF)
    if element is not None and file_to_validate is not None:
        raise ValueError(MUTUAL_EXCLUSIVE)

    try:
        schema_parser = get_schema_parser(schema)
    except AttributeError:
        return None

    if file_to_validate is not None:
        element = config.etree.parse(file_to_validate)

    try:
        schema_parser.assert_(element)  # noqa: PT009
    except AssertionError:
        log = schema_parser.error_log
        for e in log:
            try:
                parent = element.xpath(e.path)[  # type: ignore[union-attr]
                    0
                ].getparent()
            except config.etree.XPathEvalError:
                parent = element
            if parent is None:
                parent = element
            error_in_xml = config.etree.tostring(
                parent,
                encoding="UTF-8",
                pretty_print=True,
            ).decode(
                "UTF-8",
            )
            logger.error(  # noqa: TRY400
                "Error <%s> in XML:\n %s",
                e.message,
                error_in_xml,
            )
        raise
    return True



# Copyright (C) 2012-2024 Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA

"""
KML is an open standard officially named the OpenGIS KML Encoding Standard (OGC KML).

It is maintained by the Open Geospatial Consortium, Inc. (OGC).
The complete specification for OGC KML can be found at
http://www.opengeospatial.org/standards/kml/.

The complete XML schema for KML is located at
http://schemas.opengis.net/kml/.

"""
import logging
from pathlib import Path
from typing import IO
from typing import Any
from typing import AnyStr
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Union
from typing import cast

from typing_extensions import Self

from fastkml import config
from fastkml.base import _XMLObject
from fastkml.containers import Document
from fastkml.containers import Folder
from fastkml.enums import Verbosity
from fastkml.features import Placemark
from fastkml.helpers import xml_subelement_list
from fastkml.helpers import xml_subelement_list_kwarg
from fastkml.overlays import GroundOverlay
from fastkml.overlays import PhotoOverlay
from fastkml.registry import RegistryItem
from fastkml.registry import registry
from fastkml.types import Element

logger = logging.getLogger(__name__)

kml_children = Union[Folder, Document, Placemark, GroundOverlay, PhotoOverlay]


class KML(_XMLObject):
    """represents a KML File."""

    _default_nsid = config.KML

    features: List[kml_children]
    ns: str

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        features: Optional[Iterable[kml_children]] = None,
        **kwargs: Any,
    ) -> None:
        """
        Leave the namespace (ns) empty ('') if the 'kml:' prefix is undesired.

        Note that all child elements like Document or Placemark need
        to be initialized with empty namespace as well in this case.
        """
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            **kwargs,
        )
        self.features = list(features) if features is not None else []

    def __repr__(self) -> str:
        """Create a string (c)representation for KML."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"features={self.features!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    @classmethod
    def get_tag_name(cls) -> str:
        """Return the tag name."""
        return cls.__name__.lower()

    def etree_element(
        self,
        precision: Optional[int] = None,
        verbosity: Verbosity = Verbosity.normal,
    ) -> Element:
        """
        Return an Element object representing the KML element.

        Args:
        ----
            precision (Optional[int]): The precision used for floating-point values.
            verbosity (Verbosity): The verbosity level for generating the KML element.

        Returns:
        -------
            Element: The etree Element object representing the KML element.

        """
        # self.ns may be empty, which leads to unprefixed kml elements.
        # However, in this case the xlmns should still be mentioned on the kml
        # element, just without prefix.
        if not self.ns:
            root = config.etree.Element(
                f"{self.ns}{self.get_tag_name()}",
            )
            root.set("xmlns", config.KMLNS[1:-1])
        elif hasattr(config.etree, "LXML_VERSION"):
            root = config.etree.Element(
                f"{self.ns}{self.get_tag_name()}",
                nsmap={None: self.ns[1:-1]},
            )
        else:
            root = config.etree.Element(
                f"{self.ns}{self.get_tag_name()}",
            )
        xml_subelement_list(
            obj=self,
            element=root,
            attr_name="features",
            node_name="",
            precision=precision,
            verbosity=verbosity,
            default=None,
        )
        return cast(Element, root)

    def append(
        self,
        kmlobj: kml_children,
    ) -> None:
        """Append a feature."""
        self.features.append(kmlobj)

    @classmethod
    def parse(
        cls,
        file: Union[Path, str, IO[AnyStr]],
        *,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        strict: bool = True,
    ) -> Self:
        """
        Parse a KML file and return a KML object.

        Args:
        ----
            file: The file to parse.
                Can be a file path (str or Path), or a file-like object.

        Keyword Args:
        ------------
            ns (Optional[str]): The namespace of the KML file.
                If not provided, it will be inferred from the root element.
            name_spaces (Optional[Dict[str, str]]): Additional namespaces.
            strict (bool): Whether to enforce strict parsing rules. Defaults to True.

        Returns:
        -------
            KML object: The parsed KML object.

        """
        try:
            tree = config.etree.parse(
                file,
                parser=config.etree.XMLParser(
                    huge_tree=True,
                    recover=True,
                ),
            )
        except TypeError:
            tree = config.etree.parse(file)
        root = tree.getroot()
        if ns is None:
            ns = cast(str, root.tag[:-3] if root.tag.endswith("kml") else "")
        name_spaces = name_spaces or {}
        if ns:
            name_spaces["kml"] = ns
        name_spaces = {**config.NAME_SPACES, **name_spaces}
        return cls.class_from_element(
            ns=ns,
            name_spaces=name_spaces,
            strict=strict,
            element=root,
        )


registry.register(
    KML,
    RegistryItem(
        ns_ids=("kml",),
        classes=(Document, Folder, Placemark, GroundOverlay, PhotoOverlay),
        node_name="Document,Folder,Placemark,GroundOverlay,PhotoOverlay",
        attr_name="features",
        get_kwarg=xml_subelement_list_kwarg,
        set_element=xml_subelement_list,
    ),
)


__all__ = [
    "KML",
]



# Copyright (C) 2023 Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
"""Registry for XML objects."""
from dataclasses import dataclass
from typing import TYPE_CHECKING
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type

from typing_extensions import Protocol

from fastkml.enums import Verbosity
from fastkml.types import Element

if TYPE_CHECKING:
    from fastkml.base import _XMLObject


class GetKWArgs(Protocol):
    def __call__(
        self,
        *,
        element: Element,
        ns: str,
        name_spaces: Dict[str, str],
        node_name: str,
        kwarg: str,
        classes: Tuple[Type[object], ...],
        strict: bool,
    ) -> Dict[str, Any]: ...


class SetElement(Protocol):
    def __call__(
        self,
        obj: "_XMLObject",
        *,
        element: Element,
        attr_name: str,
        node_name: str,
        precision: Optional[int],
        verbosity: Verbosity,
        default: Any,
    ) -> None: ...


@dataclass(frozen=True)
class RegistryItem:
    """A registry item."""

    ns_ids: Tuple[str, ...]
    classes: Tuple[Type[object], ...]
    attr_name: str
    get_kwarg: GetKWArgs
    set_element: SetElement
    node_name: str
    default: Any = None


class Registry:
    """A registry of XML objects."""

    _registry: Dict[Type["_XMLObject"], List[RegistryItem]]

    def __init__(
        self,
        registry: Optional[Dict[Type["_XMLObject"], List[RegistryItem]]] = None,
    ) -> None:
        """Initialize the registry."""
        self._registry = registry or {}

    def __repr__(self) -> str:
        """Create a string (c)representation for Registry."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}({self._registry})"
        )

    def register(self, cls: Type["_XMLObject"], item: RegistryItem) -> None:
        """Register a class."""
        existing = self._registry.get(cls, [])
        existing.append(item)
        self._registry[cls] = existing

    def get(self, cls: Type["_XMLObject"]) -> List[RegistryItem]:
        """Get a class by name."""
        parents = reversed(cls.__mro__[:-1])
        items = []
        for parent in parents:
            items.extend(self._registry.get(parent, []))
        return items


registry = Registry()

__all__ = ["registry", "RegistryItem"]



# Copyright (C) 2012 - 2023 Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA

"""
KML Styles.

Once you've created features within Google Earth and examined the KML
code Google Earth generates, you'll notice how styles are an important
part of how your data is displayed.
"""

import logging
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Union

from fastkml import config
from fastkml.base import _XMLObject
from fastkml.enums import ColorMode
from fastkml.enums import DisplayMode
from fastkml.enums import PairKey
from fastkml.enums import Units
from fastkml.helpers import attribute_enum_kwarg
from fastkml.helpers import attribute_float_kwarg
from fastkml.helpers import bool_subelement
from fastkml.helpers import enum_attribute
from fastkml.helpers import enum_subelement
from fastkml.helpers import float_attribute
from fastkml.helpers import float_subelement
from fastkml.helpers import node_text
from fastkml.helpers import node_text_kwarg
from fastkml.helpers import subelement_bool_kwarg
from fastkml.helpers import subelement_enum_kwarg
from fastkml.helpers import subelement_float_kwarg
from fastkml.helpers import subelement_text_kwarg
from fastkml.helpers import text_subelement
from fastkml.helpers import xml_subelement
from fastkml.helpers import xml_subelement_kwarg
from fastkml.helpers import xml_subelement_list
from fastkml.helpers import xml_subelement_list_kwarg
from fastkml.kml_base import _BaseObject
from fastkml.links import Icon
from fastkml.registry import RegistryItem
from fastkml.registry import registry

logger = logging.getLogger(__name__)


class StyleUrl(_XMLObject):
    """
    URL of a <Style> or <StyleMap> defined in a Document.

    If the style is in the same file, use a # reference.
    If the style is defined in an external file,
    use a full URL along with # referencing.

    https://developers.google.com/kml/documentation/kmlreference#styleurl
    """

    _default_nsid = config.KML

    url: Optional[str]

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        url: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a Style object.

        Args:
        ----
        ns : str, optional
            The namespace of the Style object.
        name_spaces : dict, optional
            A dictionary of namespace prefixes and their corresponding URIs.
        url : str, optional
            The URL of the Style object.
        **kwargs : Any
            Additional keyword arguments.

        Returns:
        -------
        None

        """
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            **kwargs,
        )
        self.url = url

    def __repr__(self) -> str:
        """Create a string (c)representation for StyleUrl."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"url={self.url!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    def __bool__(self) -> bool:
        """
        Check if the style has a URL.

        Returns
        -------
        bool
            True if the style has a URL, False otherwise.

        """
        return bool(self.url)

    @classmethod
    def get_tag_name(cls) -> str:
        """Return the tag name."""
        return "styleUrl"


registry.register(
    StyleUrl,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="url",
        node_name="styleUrl",
        classes=(str,),
        get_kwarg=node_text_kwarg,
        set_element=node_text,
    ),
)


class _StyleSelector(_BaseObject):
    """
    Abstract element that cannot be used directly in a KML file.

    It is the base type for the <Style> and <StyleMap> elements. The
    StyleMap element selects a style based on the current mode of the
    Placemark. An element derived from StyleSelector is uniquely identified
    by its id and its url.

    https://developers.google.com/kml/documentation/kmlreference#styleselector
    """


class _ColorStyle(_BaseObject):
    """
    Abstract element that cannot be used directly in a KML file.

    It provides elements for specifying the color and color mode of
    extended style types.
    subclasses are: IconStyle, LabelStyle, LineStyle, PolyStyle.
    https://developers.google.com/kml/documentation/kmlreference#colorstyle
    """

    color: Optional[str] = None
    # Color and opacity (alpha) values are expressed in hexadecimal notation.
    # The range of values for any one color is 0 to 255 (00 to ff).
    # For alpha, 00 is fully transparent and ff is fully opaque.
    # The order of expression is aabbggrr, where aa=alpha (00 to ff);
    # bb=blue (00 to ff); gg=green (00 to ff); rr=red (00 to ff).

    color_mode: Optional[ColorMode]
    # Values for <colorMode> are normal (no effect) and random.
    # A value of random applies a random linear scale to the base <color>

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        color: Optional[str] = None,
        color_mode: Optional[ColorMode] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a Style object.

        Args:
        ----
            ns (Optional[str]): The namespace of the style.
            name_spaces (Optional[Dict[str, str]]): The dictionary of namespaces.
            id (Optional[str]): The ID of the style.
            target_id (Optional[str]): The target ID of the style.
            color (Optional[str]): The color of the style.
            color_mode (Optional[ColorMode]): The color mode of the style.
            **kwargs (Any): Additional keyword arguments.

        Returns:
        -------
            None

        """
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            id=id,
            target_id=target_id,
            **kwargs,
        )
        self.color = color
        self.color_mode = color_mode

    def __repr__(self) -> str:
        """Create a string (c)representation for _ColorStyle."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"target_id={self.target_id!r}, "
            f"color={self.color!r}, "
            f"color_mode={self.color_mode}, "
            f"**{self._get_splat()!r},"
            ")"
        )


registry.register(
    _ColorStyle,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="color",
        node_name="color",
        classes=(str,),
        get_kwarg=subelement_text_kwarg,
        set_element=text_subelement,
        default="ffffffff",
    ),
)
registry.register(
    _ColorStyle,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="color_mode",
        node_name="colorMode",
        classes=(ColorMode,),
        get_kwarg=subelement_enum_kwarg,
        set_element=enum_subelement,
        default=ColorMode.normal,
    ),
)


class HotSpot(_XMLObject):
    """
    Specifies the position within the Icon that is "anchored" to the <Point>.

    x - Either the number of pixels, a fractional component of the icon,
    or a pixel inset indicating the x component of a point on the icon.
    y - Either the number of pixels, a fractional component of the icon,
    or a pixel inset indicating the y component of a point on the icon.
    xunits - Units in which the x value is specified.
    A value of fraction indicates the x value is a fraction of the icon.
    A value of pixels indicates the x value in pixels.
    A value of insetPixels indicates the indent from the right edge of the icon.
    yunits - Units in which the y value is specified.

    https://developers.google.com/kml/documentation/kmlreference#hotspot
    """

    x: Optional[float]
    y: Optional[float]
    xunits: Optional[Units]
    yunits: Optional[Units]

    _default_nsid = config.KML

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        x: Optional[float] = None,
        y: Optional[float] = None,
        xunits: Optional[Units] = None,
        yunits: Optional[Units] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a Style element.

        Args:
        ----
            ns (Optional[str]): The namespace for the element.
            name_spaces (Optional[Dict[str, str]]): The dictionary of namespace prefixes
                and URIs.
            x (Optional[float]): The x-coordinate of the point.
            y (Optional[float]): The y-coordinate of the point.
            xunits (Optional[Units]): The units for the x-coordinate.
            yunits (Optional[Units]): The units for the y-coordinate.
            **kwargs (Any): Additional keyword arguments.

        Returns:
        -------
            None

        """
        super().__init__(ns=ns, name_spaces=name_spaces, **kwargs)
        self.x = x
        self.y = y
        self.xunits = xunits
        self.yunits = yunits

    def __repr__(self) -> str:
        """Create a string (c)representation for HotSpot."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"x={self.x!r}, "
            f"y={self.y!r}, "
            f"xunits={self.xunits}, "
            f"yunits={self.yunits}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    def __bool__(self) -> bool:
        """
        Check if both x and y attributes are not None.

        Returns
        -------
            bool: True.

        """
        return True

    @classmethod
    def get_tag_name(cls) -> str:
        """Return the tag name."""
        return "hotSpot"


registry.register(
    HotSpot,
    RegistryItem(
        ns_ids=("", "kml"),
        attr_name="x",
        node_name="x",
        classes=(float,),
        get_kwarg=attribute_float_kwarg,
        set_element=float_attribute,
        default=0.5,
    ),
)
registry.register(
    HotSpot,
    RegistryItem(
        ns_ids=("", "kml"),
        attr_name="y",
        node_name="y",
        classes=(float,),
        get_kwarg=attribute_float_kwarg,
        set_element=float_attribute,
        default=0.5,
    ),
)
registry.register(
    HotSpot,
    RegistryItem(
        ns_ids=("", "kml"),
        attr_name="xunits",
        node_name="xunits",
        classes=(Units,),
        get_kwarg=attribute_enum_kwarg,
        set_element=enum_attribute,
        default=Units.fraction,
    ),
)
registry.register(
    HotSpot,
    RegistryItem(
        ns_ids=("", "kml"),
        attr_name="yunits",
        node_name="yunits",
        classes=(Units,),
        get_kwarg=attribute_enum_kwarg,
        set_element=enum_attribute,
        default=Units.fraction,
    ),
)


class IconStyle(_ColorStyle):
    """
    Specifies how icons for point Placemarks are drawn.

    The <Icon> element specifies the icon image.
    The <scale> element specifies the x, y scaling of the icon.
    The color specified in the <color> element of <IconStyle> is blended with the
    color of the <Icon>.

    https://developers.google.com/kml/documentation/kmlreference#iconstyle
    """

    scale: Optional[float]
    # Resizes the icon. (float)
    heading: Optional[float]
    # Direction (that is, North, South, East, West), in degrees.
    # Default=0 (North).
    icon: Optional[Icon]
    # An HTTP address or a local file specification used to load an icon.
    hot_spot: Optional[HotSpot]

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        color: Optional[str] = None,
        color_mode: Optional[ColorMode] = None,
        scale: Optional[float] = None,
        heading: Optional[float] = None,
        icon: Optional[Icon] = None,
        icon_href: Optional[str] = None,
        hot_spot: Optional[HotSpot] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a Style object.

        Args:
        ----
            ns (Optional[str]): The namespace for the Style object.
            name_spaces (Optional[Dict[str, str]]): The dictionary of namespace prefixes
                and URIs.
            id (Optional[str]): The ID of the Style object.
            target_id (Optional[str]): The target ID of the Style object.
            color (Optional[str]): The color of the Style object.
            color_mode (Optional[ColorMode]): The color mode of the Style object.
            scale (Optional[float]): The scale of the Style object.
            heading (Optional[float]): The heading of the Style object.
            icon (Optional[Icon]): The icon of the Style object.
            icon_href (Optional[str]): The href of the icon can be passed as a shortcut.
            hot_spot (Optional[HotSpot]): The hot spot of the Style object.
            **kwargs (Any): Additional keyword arguments.

        Returns:
        -------
            None

        """
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            id=id,
            target_id=target_id,
            color=color,
            color_mode=color_mode,
            **kwargs,
        )
        icon_href = icon_href.strip() or None if icon_href else None
        self.scale = scale
        self.heading = heading
        if icon_href and icon:
            logger.warning(
                "Both icon_href and icon were provided. "
                "Ignoring icon_href and using icon.",
            )
        if icon_href and not icon:
            icon = Icon(ns=ns, name_spaces=name_spaces, href=icon_href)
        self.icon = icon
        self.hot_spot = hot_spot

    def __repr__(self) -> str:
        """Create a string (c)representation for IconStyle."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"target_id={self.target_id!r}, "
            f"color={self.color!r}, "
            f"color_mode={self.color_mode}, "
            f"scale={self.scale!r}, "
            f"heading={self.heading!r}, "
            f"icon={self.icon!r}, "
            f"hot_spot={self.hot_spot!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    def __bool__(self) -> bool:
        """
        Check if the Style has an icon.

        Returns
        -------
        bool
            True if the Style has an icon, False otherwise.

        """
        return bool(self.icon)

    @property
    def icon_href(self) -> Optional[str]:
        """Return the icon href."""
        return self.icon.href if self.icon else None


registry.register(
    IconStyle,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="scale",
        node_name="scale",
        classes=(float,),
        get_kwarg=subelement_float_kwarg,
        set_element=float_subelement,
        default=1.0,
    ),
)
registry.register(
    IconStyle,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="heading",
        node_name="heading",
        classes=(float,),
        get_kwarg=subelement_float_kwarg,
        set_element=float_subelement,
        default=0.0,
    ),
)
registry.register(
    IconStyle,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="icon",
        node_name="Icon",
        classes=(Icon,),
        get_kwarg=xml_subelement_kwarg,
        set_element=xml_subelement,
    ),
)
registry.register(
    IconStyle,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="hot_spot",
        node_name="hotSpot",
        classes=(HotSpot,),
        get_kwarg=xml_subelement_kwarg,
        set_element=xml_subelement,
    ),
)


class LineStyle(_ColorStyle):
    """
    The drawing style (color, color mode, and line width) for all line geometry.

    Line geometry includes the outlines of outlined polygons and the extruded "tether"
    of Placemark icons (if extrusion is enabled).
    https://developers.google.com/kml/documentation/kmlreference#linestyle
    """

    width: Optional[float]
    # Width of the line, in pixels.

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        color: Optional[str] = None,
        color_mode: Optional[ColorMode] = None,
        width: Optional[float] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a Style object.

        Args:
        ----
            ns (Optional[str]): The namespace for the Style object.
            name_spaces (Optional[Dict[str, str]]): The dictionary of namespace prefixes
                and URIs.
            id (Optional[str]): The ID of the Style object.
            target_id (Optional[str]): The target ID of the Style object.
            color (Optional[str]): The color of the Style object.
            color_mode (Optional[ColorMode]): The color mode of the Style object.
            width (Optional[float]): The width of the Style object.
            **kwargs (Any): Additional keyword arguments.

        Returns:
        -------
            None

        """
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            id=id,
            target_id=target_id,
            color=color,
            color_mode=color_mode,
            **kwargs,
        )
        self.width = width

    def __repr__(self) -> str:
        """Create a string (c)representation for LineStyle."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"target_id={self.target_id!r}, "
            f"color={self.color!r}, "
            f"color_mode={self.color_mode}, "
            f"width={self.width!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    def __bool__(self) -> bool:
        """
        Check if the style has a width defined.

        Returns
        -------
            bool: True if the width is not None, False otherwise.

        """
        return (
            self.width is not None
            or self.color is not None
            or self.color_mode is not None
        )


registry.register(
    LineStyle,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="width",
        node_name="width",
        classes=(float,),
        get_kwarg=subelement_float_kwarg,
        set_element=float_subelement,
        default=1.0,
    ),
)


class PolyStyle(_ColorStyle):
    """
    Drawing style for polygons.

    Specifies the drawing style for all polygons, including polygon
    extrusions (which look like the walls of buildings) and line
    extrusions (which look like solid fences).

    https://developers.google.com/kml/documentation/kmlreference#polystyle
    """

    fill: Optional[bool]
    # Boolean value. Specifies whether to fill the polygon.
    outline: Optional[bool]
    # Boolean value. Specifies whether to outline the polygon.
    # Polygon outlines use the current LineStyle.

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        color: Optional[str] = None,
        color_mode: Optional[ColorMode] = None,
        fill: Optional[bool] = None,
        outline: Optional[bool] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a Style object.

        Args:
        ----
            ns (Optional[str]): The namespace for the style.
            name_spaces (Optional[Dict[str, str]]): The dictionary of namespace prefixes
                and URIs.
            id (Optional[str]): The ID of the style.
            target_id (Optional[str]): The ID of the target element.
            color (Optional[str]): The color of the style.
            color_mode (Optional[ColorMode]): The color mode of the style.
            fill (Optional[bool]): Whether to fill the style.
            outline (Optional[bool]): Whether to outline the style.
            **kwargs (Any): Additional keyword arguments.

        Returns:
        -------
            None

        """
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            id=id,
            target_id=target_id,
            color=color,
            color_mode=color_mode,
            **kwargs,
        )
        self.fill = fill
        self.outline = outline

    def __repr__(self) -> str:
        """Create a string (c)representation for PolyStyle."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"target_id={self.target_id!r}, "
            f"color={self.color!r}, "
            f"color_mode={self.color_mode}, "
            f"fill={self.fill!r}, "
            f"outline={self.outline!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    def __bool__(self) -> bool:
        """
        Return True if the style has a fill or outline, False otherwise.

        Returns
        -------
            bool: True if the style has a fill or outline, False otherwise.

        """
        return (
            self.fill is not None
            or self.outline is not None
            or self.color is not None
            or self.color_mode is not None
        )


registry.register(
    PolyStyle,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="fill",
        node_name="fill",
        classes=(bool,),
        get_kwarg=subelement_bool_kwarg,
        set_element=bool_subelement,
        default=True,
    ),
)
registry.register(
    PolyStyle,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="outline",
        node_name="outline",
        classes=(bool,),
        get_kwarg=subelement_bool_kwarg,
        set_element=bool_subelement,
        default=True,
    ),
)


class LabelStyle(_ColorStyle):
    """
    Specifies how the <name> of a Feature is drawn in the 3D viewer.

    A custom color, color mode, and scale for the label (name) can be specified.

    https://developers.google.com/kml/documentation/kmlreference#labelstyle
    """

    scale: Optional[float]
    # Resizes the label.

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        color: Optional[str] = None,
        color_mode: Optional[ColorMode] = None,
        scale: Optional[float] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a Style object.

        Args:
        ----
            ns (Optional[str]): The namespace for the style.
            name_spaces (Optional[Dict[str, str]]): The dictionary of namespace prefixes
                and URIs.
            id (Optional[str]): The ID of the style.
            target_id (Optional[str]): The ID of the target element.
            color (Optional[str]): The color of the style.
            color_mode (Optional[ColorMode]): The color mode of the style.
            scale (Optional[float]): The scale of the style.
            **kwargs (Any): Additional keyword arguments.

        Returns:
        -------
            None

        """
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            id=id,
            target_id=target_id,
            color=color,
            color_mode=color_mode,
            **kwargs,
        )
        self.scale = scale

    def __repr__(self) -> str:
        """Create a string (c)representation for LabelStyle."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"target_id={self.target_id!r}, "
            f"color={self.color!r}, "
            f"color_mode={self.color_mode}, "
            f"scale={self.scale!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    def __bool__(self) -> bool:
        """
        Check if the Style has any non-None attributes.

        Returns
        -------
            bool: True if any of the following attributes are not None:
                scale, color, color_mode.

        """
        return any(
            (
                self.scale is not None,
                self.color is not None,
                self.color_mode is not None,
            ),
        )


registry.register(
    LabelStyle,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="scale",
        node_name="scale",
        classes=(float,),
        get_kwarg=subelement_float_kwarg,
        set_element=float_subelement,
        default=1.0,
    ),
)


class BalloonStyle(_BaseObject):
    """
    Specifies how the description balloon for placemarks is drawn.

    The <bgColor>, if specified, is used as the background color of the balloon.

    https://developers.google.com/kml/documentation/kmlreference#balloonstyle
    """

    bg_color: Optional[str]
    # Background color of the balloon (optional). Color and opacity (alpha)
    # values are expressed in hexadecimal notation. The range of values for
    # any one color is 0 to 255 (00 to ff). The order of expression is
    # aabbggrr, where aa=alpha (00 to ff); bb=blue (00 to ff);
    # gg=green (00 to ff); rr=red (00 to ff).
    # For alpha, 00 is fully transparent and ff is fully opaque.
    # For example, if you want to apply a blue color with 50 percent
    # opacity to an overlay, you would specify the following:
    # <bgColor>7fff0000</bgColor>, where alpha=0x7f, blue=0xff, green=0x00,
    # and red=0x00. The default is opaque white (ffffffff).
    # Note: The use of the <color> element within <BalloonStyle> has been
    # deprecated. Use <bgColor> instead.

    text_color: Optional[str]
    # Foreground color for text. The default is black (ff000000).

    text: Optional[str]
    # Text displayed in the balloon. If no text is specified, Google Earth
    # draws the default balloon (with the Feature <name> in boldface,
    # the Feature <description>, links for driving directions, a white
    # background, and a tail that is attached to the point coordinates of
    # the Feature, if specified).
    # You can add entities to the <text> tag using the following format to
    # refer to a child element of Feature: $[name], $[description], $[address],
    # $[id], $[Snippet]. Google Earth looks in the current Feature for the
    # corresponding string entity and substitutes that information in the
    # balloon.
    # To include To here - From here driving directions in the balloon,
    # use the $[geDirections] tag. To prevent the driving directions links
    # from appearing in a balloon, include the <text> element with some content
    # or with $[description] to substitute the basic Feature <description>.
    # For example, in the following KML excerpt, $[name] and $[description]
    # fields will be replaced by the <name> and <description> fields found
    # in the Feature elements that use this BalloonStyle:
    # <text>This is $[name], whose description is:<br/>$[description]</text>

    display_mode: Optional[DisplayMode]
    # If <displayMode> is default, Google Earth uses the information supplied
    # in <text> to create a balloon . If <displayMode> is hide, Google Earth
    # does not display the balloon. In Google Earth, clicking the List View
    # icon for a Placemark whose balloon's <displayMode> is hide causes
    # Google Earth to fly to the Placemark.

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        bg_color: Optional[str] = None,
        text_color: Optional[str] = None,
        text: Optional[str] = None,
        display_mode: Optional[DisplayMode] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a Style object.

        Args:
        ----
            ns (Optional[str]): The namespace for the style.
            name_spaces (Optional[Dict[str, str]]): The dictionary of namespace prefixes
            and URIs.
            id (Optional[str]): The ID of the style.
            target_id (Optional[str]): The ID of the target element.
            bg_color (Optional[str]): The background color of the style.
            text_color (Optional[str]): The text color of the style.
            text (Optional[str]): The text content of the style.
            display_mode (Optional[DisplayMode]): The display mode of the style.
            **kwargs (Any): Additional keyword arguments.

        Returns:
        -------
            None

        """
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            id=id,
            target_id=target_id,
            **kwargs,
        )
        self.bg_color = bg_color
        self.text_color = text_color
        self.text = text.strip() or None if text else None
        self.display_mode = display_mode

    def __repr__(self) -> str:
        """Create a string (c)representation for BalloonStyle."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"target_id={self.target_id!r}, "
            f"bg_color={self.bg_color!r}, "
            f"text_color={self.text_color!r}, "
            f"text={self.text!r}, "
            f"display_mode={self.display_mode}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    def __bool__(self) -> bool:
        """
        Check if the Style has any non-None attributes.

        Returns
        -------
            bool: True if any of the following attributes are not None:
                - bg_color
                - text_color
                - text
                - display_mode
            False otherwise.

        """
        return any(
            (
                self.bg_color is not None,
                self.text_color is not None,
                self.text is not None,
                self.display_mode is not None,
            ),
        )


registry.register(
    BalloonStyle,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="bg_color",
        node_name="bgColor",
        classes=(str,),
        get_kwarg=subelement_text_kwarg,
        set_element=text_subelement,
        default="ffffffff",
    ),
)
registry.register(
    BalloonStyle,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="text_color",
        node_name="textColor",
        classes=(str,),
        get_kwarg=subelement_text_kwarg,
        set_element=text_subelement,
        default="ff000000",
    ),
)
registry.register(
    BalloonStyle,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="text",
        node_name="text",
        classes=(str,),
        get_kwarg=subelement_text_kwarg,
        set_element=text_subelement,
    ),
)
registry.register(
    BalloonStyle,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="display_mode",
        node_name="displayMode",
        classes=(DisplayMode,),
        get_kwarg=subelement_enum_kwarg,
        set_element=enum_subelement,
        default=DisplayMode.default,
    ),
)


AnyStyle = Union[BalloonStyle, IconStyle, LabelStyle, LineStyle, PolyStyle]


class Style(_StyleSelector):
    """
    A Style defines an addressable style group.

    It can be referenced by StyleMaps and Features.
    Styles affect how Geometry is presented in the 3D viewer and how Features
    appear in the Places panel of the List view.
    Shared styles are collected in a <Document> and must have an id defined for them
    so that they can be referenced by the individual Features that use them.

    https://developers.google.com/kml/documentation/kmlreference#style
    """

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        styles: Optional[Iterable[AnyStyle]] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a Style element.

        Args:
        ----
            ns (Optional[str]): The namespace for the element.
            name_spaces (Optional[Dict[str, str]]): The dictionary of namespace prefixes
                and URIs.
            id (Optional[str]): The ID of the element.
            target_id (Optional[str]): The target ID of the element.
            styles (Optional[Iterable[AnyStyle]]): The list of styles.
            **kwargs (Any): Additional keyword arguments.

        Returns:
        -------
            None

        """
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            id=id,
            target_id=target_id,
            **kwargs,
        )
        self.styles = list(styles) if styles else []

    def __repr__(self) -> str:
        """Create a string (c)representation for Style."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"target_id={self.target_id!r}, "
            f"styles={self.styles!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    def __bool__(self) -> bool:
        """
        Check if the Styles object contains any styles.

        Returns
        -------
        bool
            True if the Styles object contains any styles, False otherwise.

        """
        return any(self.styles)


registry.register(
    Style,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="styles",
        node_name="Style",
        classes=(
            IconStyle,
            LabelStyle,
            LineStyle,
            PolyStyle,
            BalloonStyle,
        ),
        get_kwarg=xml_subelement_list_kwarg,
        set_element=xml_subelement_list,
    ),
)


class Pair(_BaseObject):
    """
    Stylemap pair.

    Defines a key/value pair that maps a mode (normal or highlight) to the predefined
    <styleUrl>.
    <Pair> contains two elements (both are required):
        <key>, which identifies the key
        <styleUrl> or <Style>, which references the style.
        In <styleUrl>, for referenced style elements that are local to the KML document,
        a simple # referencing is used.
        For styles that are contained in external files, use a full URL along with
        # referencing.

    https://developers.google.com/kml/documentation/kmlreference#stylemap
    """

    key: Optional[PairKey]
    style: Optional[Union[StyleUrl, Style]]

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        key: Optional[PairKey] = None,
        style: Optional[Union[StyleUrl, Style]] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a StyleSelector object.

        Args:
        ----
            ns : Optional[str]
                The namespace of the element.
            name_spaces : Optional[Dict[str, str]]
                The dictionary of namespace prefixes and URIs.
            id : Optional[str]
                The ID of the element.
            target_id : Optional[str]
                The target ID of the element.
            key : Optional[PairKey]
                The key of the element.
            style : Optional[Union[StyleUrl, Style]]
                The style or style URL of the element.
            kwargs : Any
                Additional keyword arguments.

        Returns:
        -------
         None

        """
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            id=id,
            target_id=target_id,
            **kwargs,
        )
        self.key = key
        self.style = style

    def __repr__(self) -> str:
        """Create a string (c)representation for Pair."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"target_id={self.target_id!r}, "
            f"key={self.key}, "
            f"style={self.style!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    def __bool__(self) -> bool:
        """
        Check if the StyleMap has both a key and a style.

        Returns
        -------
            bool: True if the StyleMap has both a key and a style, False otherwise.

        """
        return all((self.key is not None, self.style is not None))


registry.register(
    Pair,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="key",
        node_name="key",
        classes=(PairKey,),
        get_kwarg=subelement_enum_kwarg,
        set_element=enum_subelement,
    ),
)
registry.register(
    Pair,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="style",
        node_name="Style",
        classes=(
            StyleUrl,
            Style,
        ),
        get_kwarg=xml_subelement_kwarg,
        set_element=xml_subelement,
    ),
)


class StyleMap(_StyleSelector):
    """
    A <StyleMap> maps between two different Styles.

    Typically a <StyleMap> element is used to provide separate normal and highlighted
    styles for a placemark, so that the highlighted version appears when
    the user mouses over the icon in Google Earth.

    https://developers.google.com/kml/documentation/kmlreference#stylemap
    """

    pairs: List[Pair]

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        pairs: Optional[Iterable[Pair]] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a new Style object.

        Args:
        ----
            ns (Optional[str]): The namespace of the style.
            name_spaces (Optional[Dict[str, str]]): The dictionary of namespace prefixes
                and URIs.
            id (Optional[str]): The ID of the style.
            target_id (Optional[str]): The ID of the target element.
            pairs (Optional[Iterable[Pair]]): The list of pairs.
            **kwargs (Any): Additional keyword arguments.

        Returns:
        -------
            None

        """
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            id=id,
            target_id=target_id,
            **kwargs,
        )
        self.pairs = list(pairs) if pairs else []

    def __repr__(self) -> str:
        """Create a string (c)representation for StyleMap."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"target_id={self.target_id!r}, "
            f"pairs={self.pairs!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    def __bool__(self) -> bool:
        """
        Check if the Style has any pairs.

        Returns
        -------
        bool
            True if the Style has any pairs, False otherwise.

        """
        return bool(self.pairs)

    @property
    def normal(self) -> Optional[Union[StyleUrl, Style]]:
        """
        Get the normal style for the feature.

        Returns
        -------
            The normal style for the feature, if available. Otherwise, returns None.

        """
        return next(
            (pair.style for pair in self.pairs if pair.key == PairKey.normal),
            None,
        )

    @property
    def highlight(self) -> Optional[Union[StyleUrl, Style]]:
        """
        Return the highlight style associated with this StyleMap.

        Returns
        -------
            The highlight style, which can be either a StyleUrl or a Style object.
            If no highlight style is found, None is returned.

        """
        return next(
            (pair.style for pair in self.pairs if pair.key == PairKey.highlight),
            None,
        )


registry.register(
    StyleMap,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="pairs",
        node_name="Pair",
        classes=(Pair,),
        get_kwarg=xml_subelement_list_kwarg,
        set_element=xml_subelement_list,
    ),
)


__all__ = [
    "BalloonStyle",
    "IconStyle",
    "LabelStyle",
    "LineStyle",
    "PolyStyle",
    "Style",
    "StyleMap",
    "StyleUrl",
]



# Copyright (C) 2023  Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
"""
Feature base, Placemark and NetworkLink.

These are the objects that can be added to a KML file.
"""

import logging
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Union

from pygeoif.types import GeoCollectionType
from pygeoif.types import GeoType

from fastkml import atom
from fastkml import config
from fastkml import gx
from fastkml.base import _XMLObject
from fastkml.data import ExtendedData
from fastkml.geometry import AnyGeometryType
from fastkml.geometry import LinearRing
from fastkml.geometry import LineString
from fastkml.geometry import MultiGeometry
from fastkml.geometry import Point
from fastkml.geometry import Polygon
from fastkml.geometry import create_kml_geometry
from fastkml.helpers import attribute_int_kwarg
from fastkml.helpers import bool_subelement
from fastkml.helpers import int_attribute
from fastkml.helpers import node_text
from fastkml.helpers import node_text_kwarg
from fastkml.helpers import subelement_bool_kwarg
from fastkml.helpers import subelement_text_kwarg
from fastkml.helpers import text_subelement
from fastkml.helpers import xml_subelement
from fastkml.helpers import xml_subelement_kwarg
from fastkml.helpers import xml_subelement_list
from fastkml.helpers import xml_subelement_list_kwarg
from fastkml.kml_base import _BaseObject
from fastkml.links import Link
from fastkml.mixins import TimeMixin
from fastkml.registry import RegistryItem
from fastkml.registry import registry
from fastkml.styles import Style
from fastkml.styles import StyleMap
from fastkml.styles import StyleUrl
from fastkml.times import TimeSpan
from fastkml.times import TimeStamp
from fastkml.views import Camera
from fastkml.views import LookAt
from fastkml.views import Region

logger = logging.getLogger(__name__)

KmlGeometry = Union[
    Point,
    LineString,
    LinearRing,
    Polygon,
    MultiGeometry,
    gx.MultiTrack,
    gx.Track,
]


class Snippet(_XMLObject):
    """
    A short description of the feature.

    In Google Earth, this description is displayed in the Places panel
    under the name of the feature.
    If a Snippet is not supplied, the first two lines of the <description>
    are used.
    In Google Earth, if a Placemark contains both a description and a
    Snippet, the <Snippet> appears beneath the Placemark in the Places
    panel, and the <description> appears in the Placemark's description
    balloon.
    This tag does not support HTML markup.
    <Snippet> has a maxLines attribute, an integer that specifies the
    maximum number of lines to display.
    """

    _default_nsid = config.KML

    text: Optional[str]
    max_lines: Optional[int] = None

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        text: Optional[str] = None,
        max_lines: Optional[int] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a Snippet object.

        Args:
        ----
            ns : str, optional
                The namespace for the feature.
            name_spaces : dict[str, str], optional
                A dictionary of namespace prefixes and URIs.
            text : str, optional
                The text content of the feature.
            max_lines : int, optional
                The maximum number of lines for the feature.
            **kwargs : Any
                Additional keyword arguments.

        Returns:
        -------
        None

        """
        super().__init__(ns=ns, name_spaces=name_spaces, **kwargs)
        self.text = text.strip() or None if text else None
        self.max_lines = max_lines

    def __repr__(self) -> str:
        """
        Create a string representation for Snippet.

        Returns
        -------
            str: The string representation of the Snippet object.

        """
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"text={self.text!r}, "
            f"max_lines={self.max_lines!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    def __bool__(self) -> bool:
        """
        Check if the feature has text.

        Returns
        -------
        bool
            True if the feature has text, False otherwise.

        """
        return bool(self.text)


registry.register(
    Snippet,
    RegistryItem(
        ns_ids=("kml", ""),
        attr_name="text",
        node_name="Snippet",
        classes=(str,),
        get_kwarg=node_text_kwarg,
        set_element=node_text,
    ),
)
registry.register(
    Snippet,
    RegistryItem(
        ns_ids=("", "kml"),
        attr_name="max_lines",
        node_name="maxLines",
        classes=(int,),
        get_kwarg=attribute_int_kwarg,
        set_element=int_attribute,
        default=2,
    ),
)


class _Feature(TimeMixin, _BaseObject):
    """
    Abstract base class representing a feature in KML.

    Direct known subclasses:
        - Container (Document, Folder)
        - Placemark
        - Overlay
        - NetworkLink
        - NetworkLink.
    """

    name: Optional[str]
    visibility: Optional[bool]
    isopen: Optional[bool]
    atom_author: Optional[atom.Author]
    atom_link: Optional[atom.Link]
    address: Optional[str]
    phone_number: Optional[str]
    snippet: Optional[Snippet]
    description: Optional[str]
    style_url: Optional[StyleUrl]
    styles: List[Union[Style, StyleMap]]
    view: Union[Camera, LookAt, None]
    region: Optional[Region]
    extended_data: Optional[ExtendedData]

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        name: Optional[str] = None,
        visibility: Optional[bool] = None,
        isopen: Optional[bool] = None,
        atom_link: Optional[atom.Link] = None,
        atom_author: Optional[atom.Author] = None,
        address: Optional[str] = None,
        phone_number: Optional[str] = None,
        snippet: Optional[Snippet] = None,
        description: Optional[str] = None,
        view: Optional[Union[Camera, LookAt]] = None,
        times: Optional[Union[TimeSpan, TimeStamp]] = None,
        style_url: Optional[StyleUrl] = None,
        styles: Optional[Iterable[Union[Style, StyleMap]]] = None,
        region: Optional[Region] = None,
        extended_data: Optional[ExtendedData] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize the _Feature object.

        Args:
        ----
            ns (Optional[str]): Namespace for the element.
            name_spaces (Optional[Dict[str, str]]):
                Dictionary of namespace prefixes and URIs.
            id (Optional[str]): ID of the element.
            target_id (Optional[str]): ID of the target element.
            name (Optional[str]): User-defined text displayed in the 3D viewer as the
                label for the object.
            visibility (Optional[bool]): Specifies whether the feature is drawn in the
                3D viewer when it is initially loaded.
            isopen (Optional[bool]): Specifies whether a Document or Folder appears
                closed or open when first loaded into the Places panel.
            atom_link (Optional[atom.Link]): URL of the website containing this KML or
                KMZ file.
            atom_author (Optional[atom.Author]): Information about the author and
                related website in the KML file.
            address (Optional[str]): Unstructured address written as a standard street,
                city, state address, and/or as a postal code.
            phone_number (Optional[str]): Telephone number associated with the feature.
            snippet (Optional[Snippet]): Short description of the feature.
            description (Optional[str]): User-supplied content that appears in the
                description balloon.
            view (Optional[Union[Camera, LookAt]]): Camera or LookAt view associated.
            times (Optional[Union[TimeSpan, TimeStamp]]): Time information associated.
            style_url (Optional[StyleUrl]): URL of a Style or StyleMap.
            styles (Optional[Iterable[Union[Style, StyleMap]]]): Styles and StyleMaps
                defined to customize the appearance of the feature.
            region (Optional[Region]): Region associated with the feature.
            extended_data (Optional[ExtendedData]): Custom data added to the KML file.
            **kwargs (Any): Additional keyword arguments.

        """
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            id=id,
            target_id=target_id,
            **kwargs,
        )
        self.name = name
        self.description = description.strip() or None if description else None
        self.style_url = style_url
        self.styles = list(styles) if styles else []
        self.view = view
        self.visibility = visibility
        self.isopen = isopen
        self.snippet = snippet
        self.atom_author = atom_author
        self.atom_link = atom_link
        self.address = address
        self.phone_number = phone_number
        self.region = region
        self.extended_data = extended_data
        self.times = times

    def __repr__(self) -> str:
        """
        Return a string representation of the _Feature object.

        Returns
        -------
            str: String representation of the _Feature object.

        """
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"target_id={self.target_id!r}, "
            f"name={self.name!r}, "
            f"visibility={self.visibility!r}, "
            f"isopen={self.isopen!r}, "
            f"atom_link={self.atom_link!r}, "
            f"atom_author={self.atom_author!r}, "
            f"address={self.address!r}, "
            f"phone_number={self.phone_number!r}, "
            f"snippet={self.snippet!r}, "
            f"description={self.description!r}, "
            f"view={self.view!r}, "
            f"times={self.times!r}, "
            f"style_url={self.style_url!r}, "
            f"styles={self.styles!r}, "
            f"region={self.region!r}, "
            f"extended_data={self.extended_data!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )


registry.register(
    _Feature,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="name",
        node_name="name",
        classes=(str,),
        get_kwarg=subelement_text_kwarg,
        set_element=text_subelement,
    ),
)
registry.register(
    _Feature,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="visibility",
        node_name="visibility",
        classes=(bool,),
        get_kwarg=subelement_bool_kwarg,
        set_element=bool_subelement,
        default=True,
    ),
)
registry.register(
    _Feature,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="isopen",
        node_name="open",
        classes=(bool,),
        get_kwarg=subelement_bool_kwarg,
        set_element=bool_subelement,
        default=False,
    ),
)
registry.register(
    _Feature,
    RegistryItem(
        ns_ids=("atom",),
        attr_name="atom_author",
        node_name="atom:author",
        classes=(atom.Author,),
        get_kwarg=xml_subelement_kwarg,
        set_element=xml_subelement,
    ),
)
registry.register(
    _Feature,
    RegistryItem(
        ns_ids=("atom",),
        attr_name="atom_link",
        node_name="atom:link",
        classes=(atom.Link,),
        get_kwarg=xml_subelement_kwarg,
        set_element=xml_subelement,
    ),
)
registry.register(
    _Feature,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="address",
        node_name="address",
        classes=(str,),
        get_kwarg=subelement_text_kwarg,
        set_element=text_subelement,
    ),
)
registry.register(
    _Feature,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="phone_number",
        node_name="phoneNumber",
        classes=(str,),
        get_kwarg=subelement_text_kwarg,
        set_element=text_subelement,
    ),
)
registry.register(
    _Feature,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="snippet",
        node_name="Snippet",
        classes=(Snippet,),
        get_kwarg=xml_subelement_kwarg,
        set_element=xml_subelement,
    ),
)
registry.register(
    _Feature,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="description",
        node_name="description",
        classes=(str,),
        get_kwarg=subelement_text_kwarg,
        set_element=text_subelement,
    ),
)
registry.register(
    _Feature,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="view",
        node_name="Camera,LookAt",
        classes=(
            Camera,
            LookAt,
        ),
        get_kwarg=xml_subelement_kwarg,
        set_element=xml_subelement,
    ),
)
registry.register(
    _Feature,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="times",
        node_name="TimeSpan,TimeStamp",
        classes=(
            TimeSpan,
            TimeStamp,
        ),
        get_kwarg=xml_subelement_kwarg,
        set_element=xml_subelement,
    ),
)
registry.register(
    _Feature,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="style_url",
        node_name="styleUrl",
        classes=(StyleUrl,),
        get_kwarg=xml_subelement_kwarg,
        set_element=xml_subelement,
    ),
)
registry.register(
    _Feature,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="styles",
        node_name="Style,StyleMap",
        classes=(
            Style,
            StyleMap,
        ),
        get_kwarg=xml_subelement_list_kwarg,
        set_element=xml_subelement_list,
    ),
)
registry.register(
    _Feature,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="region",
        node_name="region",
        classes=(Region,),
        get_kwarg=xml_subelement_kwarg,
        set_element=xml_subelement,
    ),
)
registry.register(
    _Feature,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="extended_data",
        node_name="ExtendedData",
        classes=(ExtendedData,),
        get_kwarg=xml_subelement_kwarg,
        set_element=xml_subelement,
    ),
)


class Placemark(_Feature):
    """
    A Placemark is a Feature with associated Geometry.

    In Google Earth, a Placemark appears as a list item in the Places
    panel. A Placemark with a Point has an icon associated with it that
    marks a point on the Earth in the 3D viewer.
    """

    kml_geometry: Optional[KmlGeometry]

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        name: Optional[str] = None,
        visibility: Optional[bool] = None,
        isopen: Optional[bool] = None,
        atom_link: Optional[atom.Link] = None,
        atom_author: Optional[atom.Author] = None,
        address: Optional[str] = None,
        phone_number: Optional[str] = None,
        snippet: Optional[Snippet] = None,
        description: Optional[str] = None,
        view: Optional[Union[Camera, LookAt]] = None,
        times: Optional[Union[TimeSpan, TimeStamp]] = None,
        style_url: Optional[StyleUrl] = None,
        styles: Optional[Iterable[Union[Style, StyleMap]]] = None,
        region: Optional[Region] = None,
        extended_data: Optional[ExtendedData] = None,
        # Placemark specific
        kml_geometry: Optional[KmlGeometry] = None,
        geometry: Optional[Union[GeoType, GeoCollectionType]] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a Feature object.

        Parameters
        ----------
        ns : str, optional
            The namespace for the feature.
        name_spaces : dict[str, str], optional
            The dictionary of namespace prefixes and URIs.
        id : str, optional
            The ID of the feature.
        target_id : str, optional
            The target ID of the feature.
        name : str, optional
            The name of the feature.
        visibility : bool, optional
            The visibility of the feature.
        isopen : bool, optional
            The open status of the feature.
        atom_link : atom.Link, optional
            The Atom link associated with the feature.
        atom_author : atom.Author, optional
            The Atom author associated with the feature.
        address : str, optional
            The address of the feature.
        phone_number : str, optional
            The phone number of the feature.
        snippet : Snippet, optional
            The snippet of the feature.
        description : str, optional
            The description of the feature.
        view : Union[Camera, LookAt], optional
            The view associated with the feature.
        times : Union[TimeSpan, TimeStamp], optional
            The times associated with the feature.
        style_url : StyleUrl, optional
            The style URL of the feature.
        styles : Iterable[Union[Style, StyleMap]], optional
            The styles associated with the feature.
        region : Region, optional
            The region associated with the feature.
        extended_data : ExtendedData, optional
            The extended data associated with the feature.
        kml_geometry : KmlGeometry, optional
            The KML geometry associated with the feature.
        geometry : Union[GeoType, GeoCollectionType], optional
            The geometry associated with the feature.
        **kwargs : Any
            Additional keyword arguments.

        Raises
        ------
        ValueError
            If both `kml_geometry` and `geometry` are specified.

        """
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            id=id,
            target_id=target_id,
            name=name,
            visibility=visibility,
            isopen=isopen,
            atom_link=atom_link,
            atom_author=atom_author,
            address=address,
            phone_number=phone_number,
            snippet=snippet,
            description=description,
            view=view,
            times=times,
            style_url=style_url,
            styles=styles,
            region=region,
            extended_data=extended_data,
            **kwargs,
        )
        if kml_geometry and geometry:
            msg = "You can only specify one of kml_geometry or geometry"
            raise ValueError(msg)
        if geometry:
            kml_geometry = create_kml_geometry(
                geometry=geometry,
                ns=ns,
                name_spaces=name_spaces,
            )
        self.kml_geometry = kml_geometry

    def __repr__(self) -> str:
        """
        Return a string representation of the Placemark object.

        Returns
        -------
            str: A string representation of the Placemark object.

        """
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"target_id={self.target_id!r}, "
            f"name={self.name!r}, "
            f"visibility={self.visibility!r}, "
            f"isopen={self.isopen!r}, "
            f"atom_link={self.atom_link!r}, "
            f"atom_author={self.atom_author!r}, "
            f"address={self.address!r}, "
            f"phone_number={self.phone_number!r}, "
            f"snippet={self.snippet!r}, "
            f"description={self.description!r}, "
            f"view={self.view!r}, "
            f"times={self.times!r}, "
            f"style_url={self.style_url!r}, "
            f"styles={self.styles!r}, "
            f"region={self.region!r}, "
            f"extended_data={self.extended_data!r}, "
            f"kml_geometry={self.kml_geometry!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    @property
    def geometry(self) -> Optional[AnyGeometryType]:
        """
        Returns the geometry associated with this feature.

        Returns
        -------
            Optional[AnyGeometryType]: The geometry associated with this feature,
            or None if no geometry is present.

        """
        return self.kml_geometry.geometry if self.kml_geometry is not None else None


registry.register(
    Placemark,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="kml_geometry",
        node_name=(
            "Point,LineString,LinearRing,Polygon,MultiGeometry,"
            "gx:MultiTrack,gx:Track"
        ),
        classes=(
            Point,
            LineString,
            LinearRing,
            Polygon,
            MultiGeometry,
            gx.MultiTrack,
            gx.Track,
        ),
        get_kwarg=xml_subelement_kwarg,
        set_element=xml_subelement,
    ),
)


class NetworkLink(_Feature):
    """
    References a KML file or KMZ archive on a local or remote network.

    Use the <Link> element to specify the location of the KML file.
    Within that element, you can define the refresh options for updating the file,
    based on time and camera change.
    NetworkLinks can be used in combination with Regions to handle very large datasets
    efficiently.
    https://developers.google.com/kml/documentation/kmlreference#networklink


    Elements Specific to NetworkLink
    <refreshVisibility>
    Boolean value.
    A value of 0 leaves the visibility of features within the control of the
    Google Earth user.
    Set the value to 1 to reset the visibility of features each time the NetworkLink is
    refreshed.
    For example, suppose a Placemark within the linked KML file has <visibility>
    set to 1 and the NetworkLink has <refreshVisibility> set to 1.
    When the file is first loaded into Google Earth, the user can clear the check box
    next to the item to turn off display in the 3D viewer.
    However, when the NetworkLink is refreshed, the Placemark will be made
    visible again, since its original visibility state was TRUE.
    <flyToView>
    Boolean value.
    A value of 1 causes Google Earth to fly to the view of the LookAt or Camera in the
    NetworkLinkControl (if it exists).
    If the NetworkLinkControl does not contain an AbstractView element,
    Google Earth flies to the LookAt or Camera element in the Feature child
    within the <kml> element in the refreshed file.
    If the <kml> element does not have a LookAt or Camera specified,
    the view is unchanged.
    For example, Google Earth would fly to the <LookAt> view of the parent Document,
    not the <LookAt> of the Placemarks contained within the Document.
    <Link>(required)

    https://developers.google.com/kml/documentation/kmlreference#networklink
    """

    refresh_visibility: Optional[bool]
    fly_to_view: Optional[bool]
    link: Optional[Link]

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        name: Optional[str] = None,
        visibility: Optional[bool] = None,
        isopen: Optional[bool] = None,
        atom_link: Optional[atom.Link] = None,
        atom_author: Optional[atom.Author] = None,
        address: Optional[str] = None,
        phone_number: Optional[str] = None,
        snippet: Optional[Snippet] = None,
        description: Optional[str] = None,
        view: Optional[Union[Camera, LookAt]] = None,
        times: Optional[Union[TimeSpan, TimeStamp]] = None,
        style_url: Optional[StyleUrl] = None,
        styles: Optional[Iterable[Union[Style, StyleMap]]] = None,
        region: Optional[Region] = None,
        extended_data: Optional[ExtendedData] = None,
        # NetworkLink specific
        refresh_visibility: Optional[bool] = None,
        fly_to_view: Optional[bool] = None,
        link: Optional[Link] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a Feature object.

        Parameters
        ----------
        ns : str, optional
            The namespace for the feature.
        name_spaces : dict[str, str], optional
            The dictionary of namespace prefixes and URIs.
        id : str, optional
            The ID of the feature.
        target_id : str, optional
            The target ID of the feature.
        name : str, optional
            The name of the feature.
        visibility : bool, optional
            The visibility of the feature.
        isopen : bool, optional
            Whether the feature is open.
        atom_link : atom.Link, optional
            The Atom link associated with the feature.
        atom_author : atom.Author, optional
            The Atom author associated with the feature.
        address : str, optional
            The address of the feature.
        phone_number : str, optional
            The phone number of the feature.
        snippet : Snippet, optional
            The snippet associated with the feature.
        description : str, optional
            The description of the feature.
        view : Union[Camera, LookAt], optional
            The view associated with the feature.
        times : Union[TimeSpan, TimeStamp], optional
            The times associated with the feature.
        style_url : StyleUrl, optional
            The style URL of the feature.
        styles : Iterable[Union[Style, StyleMap]], optional
            The styles associated with the feature.
        region : Region, optional
            The region associated with the feature.
        extended_data : ExtendedData, optional
            The extended data associated with the feature.
        refresh_visibility : bool, optional
            The refresh visibility of the feature (NetworkLink specific).
        fly_to_view : bool, optional
            Whether to fly to the view (NetworkLink specific).
        link : Link, optional
            The link associated with the feature.
        **kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        None

        """
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            id=id,
            target_id=target_id,
            name=name,
            visibility=visibility,
            isopen=isopen,
            atom_link=atom_link,
            atom_author=atom_author,
            address=address,
            phone_number=phone_number,
            snippet=snippet,
            description=description,
            view=view,
            times=times,
            style_url=style_url,
            styles=styles,
            region=region,
            extended_data=extended_data,
            **kwargs,
        )
        self.refresh_visibility = refresh_visibility
        self.fly_to_view = fly_to_view
        self.link = link

    def __repr__(self) -> str:
        """
        Return a string representation of the NetworkLink object.

        Returns
        -------
            str: A string representation of the NetworkLink object.

        """
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"target_id={self.target_id!r}, "
            f"name={self.name!r}, "
            f"visibility={self.visibility!r}, "
            f"isopen={self.isopen!r}, "
            f"atom_link={self.atom_link!r}, "
            f"atom_author={self.atom_author!r}, "
            f"address={self.address!r}, "
            f"phone_number={self.phone_number!r}, "
            f"snippet={self.snippet!r}, "
            f"description={self.description!r}, "
            f"view={self.view!r}, "
            f"times={self.times!r}, "
            f"style_url={self.style_url!r}, "
            f"styles={self.styles!r}, "
            f"region={self.region!r}, "
            f"extended_data={self.extended_data!r}, "
            f"refresh_visibility={self.refresh_visibility!r}, "
            f"fly_to_view={self.fly_to_view!r}, "
            f"link={self.link!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    def __bool__(self) -> bool:
        """
        Check if the feature has a link.

        Returns
        -------
        bool
            True if the feature has a link, False otherwise.

        """
        return bool(self.link)


registry.register(
    NetworkLink,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="refresh_visibility",
        node_name="refreshVisibility",
        classes=(bool,),
        get_kwarg=subelement_bool_kwarg,
        set_element=bool_subelement,
        default=False,
    ),
)
registry.register(
    NetworkLink,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="fly_to_view",
        node_name="flyToView",
        classes=(bool,),
        get_kwarg=subelement_bool_kwarg,
        set_element=bool_subelement,
        default=False,
    ),
)
registry.register(
    NetworkLink,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="link",
        node_name="Link",
        classes=(Link,),
        get_kwarg=xml_subelement_kwarg,
        set_element=xml_subelement,
    ),
)
